// FreeTypeTest3.cpp
// Copyright (C) 2017 Katayama Hirofumi MZ.
// This software is public domain software (PDS).
//////////////////////////////////////////////////////////////////////////////

#include <ft2build.h>
#include FT_FREETYPE_H
#include FT_GLYPH_H

#include "MWindowBase.hpp"
#include <cmath>

FT_Library library;
FT_Face face;
const FT_UInt g_pixel_height = 100;

//////////////////////////////////////////////////////////////////////////////

typedef struct BITMAPINFODX
{
    BITMAPINFOHEADER    bmiHeader;
    RGBQUAD             bmiColors[256];
} BITMAPINFODX;

BOOL SaveBitmapToFile(LPCTSTR pszFileName, HBITMAP hbm)
{
    BOOL f;
    BITMAPFILEHEADER bf;
    BITMAPINFODX bi;
    BITMAPINFOHEADER *pbmih;
    DWORD cb;
    DWORD cColors, cbColors;
    HDC hDC;
    HANDLE hFile;
    LPVOID pBits;
    BITMAP bm;

    if (!GetObject(hbm, sizeof(BITMAP), &bm))
        return FALSE;

    pbmih = &bi.bmiHeader;
    ZeroMemory(pbmih, sizeof(BITMAPINFOHEADER));
    pbmih->biSize             = sizeof(BITMAPINFOHEADER);
    pbmih->biWidth            = bm.bmWidth;
    pbmih->biHeight           = bm.bmHeight;
    pbmih->biPlanes           = 1;
    pbmih->biBitCount         = bm.bmBitsPixel;
    pbmih->biCompression      = BI_RGB;
    pbmih->biSizeImage        = bm.bmWidthBytes * bm.bmHeight;

    if (bm.bmBitsPixel < 16)
        cColors = 1 << bm.bmBitsPixel;
    else
        cColors = 0;
    cbColors = cColors * sizeof(RGBQUAD);

    bf.bfType = 0x4d42;
    bf.bfReserved1 = 0;
    bf.bfReserved2 = 0;
    cb = sizeof(BITMAPFILEHEADER) + pbmih->biSize + cbColors;
    bf.bfOffBits = cb;
    bf.bfSize = cb + pbmih->biSizeImage;

    pBits = HeapAlloc(GetProcessHeap(), 0, pbmih->biSizeImage);
    if (pBits == NULL)
        return FALSE;

    f = FALSE;
    hDC = GetDC(NULL);
    if (hDC != NULL)
    {
        if (GetDIBits(hDC, hbm, 0, bm.bmHeight, pBits, (BITMAPINFO*)&bi,
            DIB_RGB_COLORS))
        {
            hFile = CreateFile(pszFileName, GENERIC_WRITE, FILE_SHARE_READ, NULL,
                               CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL |
                               FILE_FLAG_WRITE_THROUGH, NULL);
            if (hFile != INVALID_HANDLE_VALUE)
            {
                f = WriteFile(hFile, &bf, sizeof(BITMAPFILEHEADER), &cb, NULL) &&
                    WriteFile(hFile, &bi, sizeof(BITMAPINFOHEADER), &cb, NULL) &&
                    WriteFile(hFile, bi.bmiColors, cbColors, &cb, NULL) &&
                    WriteFile(hFile, pBits, pbmih->biSizeImage, &cb, NULL);
                CloseHandle(hFile);

                if (!f)
                    DeleteFile(pszFileName);
            }
        }
        ReleaseDC(NULL, hDC);
    }

    HeapFree(GetProcessHeap(), 0, pBits);
    return f;
}

HBITMAP ConvertToGrayscaleBitmap(FT_Bitmap bitmap)
{
    HBITMAP hbm;
    BITMAPINFODX bi;
    
    ZeroMemory(&bi, sizeof(bi));
    bi.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
    bi.bmiHeader.biWidth = bitmap.width;
    bi.bmiHeader.biHeight = -bitmap.rows;
    bi.bmiHeader.biPlanes = 1;
    bi.bmiHeader.biBitCount = 8;
    bi.bmiHeader.biCompression = BI_RGB;
    
    for (INT i = 0; i < 256; ++i)
    {
        bi.bmiColors[i].rgbBlue = (BYTE)i;
        bi.bmiColors[i].rgbGreen = (BYTE)i;
        bi.bmiColors[i].rgbRed = (BYTE)i;
        bi.bmiColors[i].rgbReserved = 0;
    }

    LPVOID pvBits;
    hbm = CreateDIBSection(NULL, (BITMAPINFO *)&bi, DIB_RGB_COLORS, &pvBits, NULL, 0);
    if (hbm == NULL)
        return NULL;

    BITMAP bm;
    GetObject(hbm, sizeof(bm), &bm);

    LPBYTE pbBits = (LPBYTE)pvBits;
    for (INT y = 0; y < bitmap.rows; ++y)
    {
        CopyMemory(&pbBits[y * bm.bmWidthBytes], 
                   &bitmap.buffer[y * bitmap.pitch],
                   bitmap.width);
    }

    return hbm;
}

//////////////////////////////////////////////////////////////////////////////

struct TransMatrix
{
    FT_Matrix matrix;
    FT_Vector vector;
};

void TransMatrix_Multiply(const TransMatrix *a, TransMatrix *b)
{
    FT_Matrix *bm = &b->matrix;
    FT_Pos ax = a->vector.x;
    FT_Pos ay = a->vector.y;
    FT_Pos rx = FT_MulFix(ax, bm->xx) + FT_MulFix(ay, bm->yx) + b->vector.x;
    FT_Pos ry = FT_MulFix(ax, bm->xy) + FT_MulFix(ay, bm->yy) + b->vector.y;
    FT_Matrix_Multiply(&a->matrix, bm);
    b->vector.x = rx;
    b->vector.y = ry;
}

FT_Error LoadFace(const char *pathname)
{
    return FT_New_Face(library, pathname, 0, &face);
}

TransMatrix GetTransMatrix(INT i)
{
    double radian = (10 * i * M_PI) / 180;

    TransMatrix tmat;
    tmat.matrix.xx = FT_Fixed(std::cos(radian) * (1 << 16));
    tmat.matrix.xy = FT_Fixed(-std::sin(radian) * (1 << 16));
    tmat.matrix.yx = FT_Fixed(std::sin(radian) * (1 << 16));
    tmat.matrix.yy = FT_Fixed(std::cos(radian) * (1 << 16));
    tmat.vector.x = 0;
    tmat.vector.y = 0;

    FT_Vector vec;
    vec.x = 0;
    vec.y = (face->ascender * face->size->metrics.y_ppem / face->units_per_EM) << 6;
    FT_Vector_Transform(&vec, &tmat.matrix);

    tmat.vector.x = -vec.x;
    tmat.vector.y = -vec.y;

    return tmat;
}

FT_GlyphSlot CreateGlyphSlot(WCHAR charcode, INT i)
{
    FT_Error error;

    error = FT_Set_Pixel_Sizes(face, 0, g_pixel_height);
    if (error)
    {
        DebugPrintDx("FT_Set_Pixel_Sizes\n");
        return NULL;
    }

    TransMatrix tmat = GetTransMatrix(i);
    FT_Set_Transform(face, &tmat.matrix, &tmat.vector);

    FT_UInt glyph_index = FT_Get_Char_Index(face, charcode);
    FT_Int32 load_flags = FT_LOAD_DEFAULT;
    error = FT_Load_Glyph(face, glyph_index, load_flags);
    if (error)
    {
        DebugPrintDx("FT_Load_Glyph\n");
        return NULL;
    }

    FT_Render_Mode render_mode = FT_RENDER_MODE_NORMAL;
    error = FT_Render_Glyph(face->glyph, render_mode);
    if (error)
    {
        DebugPrintDx("FT_Render_Glyph\n");
        return NULL;
    }

    return face->glyph;
}

//////////////////////////////////////////////////////////////////////////////

struct MFreeTypeTest3 : public MWindowBase
{
    HINSTANCE   m_hInst;        // the instance handle
    INT         m_i;

    // constructors
    MFreeTypeTest3(HINSTANCE hInst) : m_hInst(hInst), m_i(0)
    {
    }

    virtual void ModifyWndClassDx(WNDCLASSEX& wcx)
    {
        wcx.hbrBackground = CreateSolidBrush(RGB(255, 255, 0));
    }

    BOOL StartDx(INT nCmdShow)
    {
        if (!CreateWindowDx(NULL, TEXT("FreeTypeTest3"),
                            WS_OVERLAPPEDWINDOW, 0,
                            CW_USEDEFAULT, CW_USEDEFAULT, 600, 500))
        {
            MsgBoxDx(TEXT("failure of CreateWindow"), NULL, MB_ICONERROR);
            return FALSE;
        }

        ::ShowWindow(*this, nCmdShow);
        ::UpdateWindow(*this);

        return TRUE;
    }

    // message loop
    INT RunDx()
    {
        MSG msg;
        while (::GetMessage(&msg, NULL, 0, 0))
        {
            ::TranslateMessage(&msg);
            ::DispatchMessage(&msg);
        }
        return INT(msg.wParam);
    }

    BOOL OnCreate(HWND hwnd, LPCREATESTRUCT lpCreateStruct)
    {
        SetTimer(hwnd, 999, 500, NULL);
        return TRUE;
    }

    void OnTimer(HWND hwnd, UINT id)
    {
        m_i++;
        InvalidateRect(hwnd, NULL, TRUE);
    }

    void OnDestroy(HWND hwnd)
    {
        KillTimer(hwnd, 999);
        ::PostQuitMessage(0);
    }

    void DoPaint(HWND hwnd, HDC hDC)
    {
        RECT rc;
        GetClientRect(hwnd, &rc);

        // center point of the window
        POINT pt;
        pt.x = (rc.left + rc.right) / 2;
        pt.y = (rc.top + rc.bottom) / 2;

        WCHAR str[] = {'A', 'B', 'C', ',' };

        // get width
        INT dx = 0;
        for (size_t i = 0; i < _countof(str); ++i)
        {
            FT_GlyphSlot slot = CreateGlyphSlot(str[i], 0);
            dx += (slot->advance.x >> 6);
        }

        // calculate background box
        INT ty = 0;
        INT hy = -face->height * face->size->metrics.y_ppem / face->units_per_EM;
        TransMatrix tmat = GetTransMatrix(m_i);
        FT_Vector vecs[4];
        POINT pts[4];
        vecs[0].x = 0;
        vecs[0].y = hy << 16;
        vecs[1].x = dx << 16;
        vecs[1].y = hy << 16;
        vecs[2].x = dx << 16;
        vecs[2].y = ty << 16;
        vecs[3].x = 0;
        vecs[3].y = ty << 16;
        for (int i = 0; i < 4; ++i)
        {
            FT_Vector_Transform(&vecs[i], &tmat.matrix);
            pts[i].x = pt.x + (vecs[i].x >> 16);
            pts[i].y = pt.y - (vecs[i].y >> 16);
        }

        // fill background box
        HBRUSH hbr = CreateSolidBrush(RGB(0, 255, 0));
        HGDIOBJ hbrOld = SelectObject(hDC, hbr);
        HGDIOBJ hPenOld = SelectObject(hDC, GetStockObject(WHITE_PEN));
        Polygon(hDC, pts, 4);
        SelectObject(hDC, hPenOld);
        SelectObject(hDC, hbrOld);
        DeleteObject(hbr);

        // draw glyphs
        INT x, y;
        x = pt.x;
        y = pt.y;
        HDC hdcMem = CreateCompatibleDC(hDC);
        for (size_t i = 0; i < _countof(str); ++i)
        {
            FT_GlyphSlot slot = CreateGlyphSlot(str[i], m_i);
            FT_Bitmap bitmap = slot->bitmap;

            HBITMAP hbm = ConvertToGrayscaleBitmap(bitmap);
            HGDIOBJ hbmOld = SelectObject(hdcMem, hbm);
            {
                INT rx = x + slot->bitmap_left;
                INT ry = y - slot->bitmap_top;
                BitBlt(hDC, rx, ry,
                       bitmap.width, bitmap.rows, hdcMem, 0, 0, SRCCOPY);
            }
            SelectObject(hdcMem, hbmOld);
            DeleteObject(hbm);

            x += (slot->advance.x >> 6);
            y -= (slot->advance.y >> 6);
        }
        DeleteDC(hdcMem);

        // draw the original top line
        MoveToEx(hDC, rc.left, pt.y, NULL);
        LineTo(hDC, rc.right, pt.y);

        // draw the rotated top line
        vecs[0].x = -(pt.x * 2) << 16;
        vecs[0].y = 0;
        vecs[1].x = (pt.x * 2) << 16;
        vecs[1].y = 0;
        for (int i = 0; i < 2; ++i)
        {
            FT_Vector_Transform(&vecs[i], &tmat.matrix);
            pts[i].x = pt.x + (vecs[i].x >> 16);
            pts[i].y = pt.y - (vecs[i].y >> 16);
        }
        HPEN hPen = CreatePen(PS_SOLID, 1, RGB(0, 0, 255));
        hPenOld = SelectObject(hDC, hPen);
        Polyline(hDC, pts, 2);
        SelectObject(hDC, hPenOld);
        DeleteObject(hPen);
    }

    void OnPaint(HWND hwnd)
    {
        RECT rc;
        ::GetClientRect(m_hwnd, &rc);

        PAINTSTRUCT ps;
        HDC hDC = ::BeginPaint(m_hwnd, &ps);
        if (hDC)
        {
            DoPaint(hwnd, hDC);

            ::EndPaint(m_hwnd, &ps);
        }
    }

    void OnSize(HWND hwnd, UINT state, int cx, int cy)
    {
        ::InvalidateRect(m_hwnd, NULL, TRUE);
    }

    virtual LRESULT CALLBACK
    WindowProcDx(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
    {
        switch (uMsg)
        {
            DO_MSG(WM_CREATE, OnCreate);
            DO_MSG(WM_TIMER, OnTimer);
            DO_MSG(WM_SIZE, OnSize);
            DO_MSG(WM_PAINT, OnPaint);
            DO_MSG(WM_DESTROY, OnDestroy);
        default:
            return DefaultProcDx(hwnd, uMsg, wParam, lParam);
        }
        return 0;
    }
};

const char *GetFontPath(const char *name)
{
    static char path[MAX_PATH];
    GetWindowsDirectoryA(path, MAX_PATH);
    lstrcat(path, "\\Fonts\\");
    lstrcat(path, name);
    return path;
}

//////////////////////////////////////////////////////////////////////////////
// Win32 App main function

extern "C" INT WINAPI
_tWinMain(HINSTANCE   hInstance,
          HINSTANCE   hPrevInstance,
          LPTSTR      lpCmdLine,
          INT         nCmdShow)
{
    int ret;

    FT_Error error;

    error = FT_Init_FreeType(&library);
    if (error)
    {
        DebugPrintDx("FT_Init_FreeType\n");
        return 4;
    }

    error = LoadFace(GetFontPath("arial.ttf"));
    if (error)
    {
        DebugPrintDx("FT_New_Face\n");
        return 3;
    }
    DebugPrintDx("family_name: %s\n", face->family_name);

    {
        MFreeTypeTest3 app(GetModuleHandle(NULL));

        if (app.StartDx(nCmdShow))
        {
            ret = app.RunDx();
        }
        else
        {
            ret = 2;
        }
    }

    FT_Done_Face(face);
    FT_Done_FreeType(library);

    return ret;
}

//////////////////////////////////////////////////////////////////////////////
